@@ -1,9 +1,28 @@
+/*  GRT stacks implementation for linux and other *nix.
+    Copyright (C) 2002, 2003, 2004, 2005 Tristan Gingold.
+
+    GHDL is free software; you can redistribute it and/or modify it under
+    the terms of the GNU General Public License as published by the Free
+    Software Foundation; either version 2, or (at your option) any later
+    version.
+
+    GHDL is distributed in the hope that it will be useful, but WITHOUT ANY
+    WARRANTY; without even the implied warranty of MERCHANTABILITY or
+    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
+    for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with GCC; see the file COPYING.  If not, write to the Free
+    Software Foundation, 59 Temple Place - Suite 330, Boston, MA
+    02111-1307, USA.
+*/
 #define _GNU_SOURCE
 #include <unistd.h>
 #include <sys/mman.h>
 #include <signal.h>
 #include <fcntl.h>
 #include <sys/ucontext.h>
+#include <stdlib.h>
 //#include <stdint.h>
 
 /* On x86, the stack growns downward.  */
@@ -43,11 +62,6 @@ struct stack_context
   size_t cur_length;
 };
 
-/* Context for the main stack.  */
-static struct stack_context main_stack_context;
-
-extern struct stack_context *grt_stack_main_stack;
-
 /* If MAP_ANONYMOUS is not defined, use /dev/zero. */
 #ifndef MAP_ANONYMOUS
 #define USE_DEV_ZERO
@@ -59,10 +73,8 @@ static int dev_zero_fd;
 #endif
 
 #if EXTEND_STACK
-/* Defined in Grt.Processes (body).
-   This is the current process being run.
-   FIXME: this won't work with pthread!  */
-extern void **grt_cur_proc;
+/* This is the current process being run.  */
+extern struct stack_context *grt_get_current_process (void);
 
 /* Stack used for signals.
    The stack must be different from the running stack, because we want to be
@@ -105,7 +117,7 @@ static void grt_sigsegv_handler (int signo, siginfo_t *info, void *ptr)
     }
 #endif
 
-  if (info == NULL || grt_cur_proc == NULL || in_handler > 1)
+  if (info == NULL || grt_get_current_process () == NULL || in_handler > 1)
     {
       /* We loose.  */
       sigaction (SIGSEGV, &prev_sigsegv_act, NULL);
@@ -115,7 +127,7 @@ static void grt_sigsegv_handler (int signo, siginfo_t *info, void *ptr)
   addr = info->si_addr;
   
   /* Check ADDR belong to the stack.  */
-  ctxt = *grt_cur_proc;
+  ctxt = grt_get_current_process ()->cur_sp;
   stack_high = (void *)(ctxt + 1);
   stack_low = stack_high - stack_max_size;
   if (addr > stack_high || addr < stack_low)
@@ -176,6 +188,19 @@ static void grt_signal_setup (void)
 }
 #endif
 
+/* Context for the main stack.  */
+static __thread struct stack_context main_stack_context;
+
+extern void grt_set_main_stack (struct stack_context *stack);
+
+void
+grt_stack_new_thread (void)
+{
+  main_stack_context.cur_sp = NULL;
+  main_stack_context.cur_length = 0;
+  grt_set_main_stack (&main_stack_context);
+}
+
 void
 grt_stack_init (void)
 {
@@ -197,7 +222,7 @@ grt_stack_init (void)
   /* Initialize the main stack context.  */
   main_stack_context.cur_sp = NULL;
   main_stack_context.cur_length = 0;
-  grt_stack_main_stack = &main_stack_context;
+  grt_set_main_stack (&main_stack_context);
 
 #ifdef USE_DEV_ZERO
   dev_zero_fd = open ("/dev/zero", O_RDWR);
@@ -244,6 +269,13 @@ grt_stack_allocate (void)
   res = (struct stack_context *)(base + sizeof (struct stack_context));
 #endif
 
+#ifdef __ia64__
+  /* Also allocate BSP.  */
+  if (mmap (base, page_size, PROT_READ | PROT_WRITE,
+	    MAP_FIXED | MAP_PRIVATE | MAP_ANONYMOUS, MMAP_FILEDES, 0) != base)
+    return NULL;
+#endif
+
   res->cur_sp = (void *)res;
   res->cur_length = stack_size;
   return res;
